# Copyright 2018 ACSONE SA/NV (<http://acsone.eu>)
# License LGPL-3.0 or later (http://www.gnu.org/licenses/lgpl.html).

import logging
from contextlib import contextmanager

try:
    import odoo
    from odoo.api import Environment

    odoo_bin = "odoo"
except ImportError:
    # Odoo < 10
    try:
        import openerp as odoo
        from openerp.api import Environment

        odoo_bin = "openerp-server"
    except ImportError:
        raise ImportError("No module named odoo nor openerp")

_logger = logging.getLogger(__name__)


@contextmanager
def OdooEnvironment(database, rollback=False, **kwargs):
    with Environment.manage():
        registry = odoo.registry(database)
        try:
            with registry.cursor() as cr:
                uid = odoo.SUPERUSER_ID
                try:
                    ctx = Environment(cr, uid, {})["res.users"].context_get()
                except Exception as e:
                    ctx = {"lang": "en_US"}
                    # this happens, for instance, when there are new
                    # fields declared on res_partner which are not yet
                    # in the database (before -u)
                    _logger.warning(
                        "Could not obtain a user context, continuing "
                        "anyway with a default context. Error was: %s",
                        e,
                    )
                env = Environment(cr, uid, ctx)
                cr.rollback()
                yield env
                if rollback:
                    cr.rollback()
                else:
                    cr.commit()
        finally:
            if odoo.tools.parse_version(
                odoo.release.version
            ) < odoo.tools.parse_version("10.0"):
                odoo.modules.registry.RegistryManager.delete(database)
            else:
                odoo.modules.registry.Registry.delete(database)
            odoo.sql_db.close_db(database)



# Copyright 2018 ACSONE SA/NV (<http://acsone.eu>)
# License LGPL-3.0 or later (http://www.gnu.org/licenses/lgpl.html).

import code
import logging
import os

_logger = logging.getLogger(__name__)


class Shell(object):

    shells = ["ipython", "ptpython", "bpython", "python"]

    @classmethod
    def python(cls, local_vars):
        console = code.InteractiveConsole(locals=local_vars)
        import rlcompleter  # noqa
        import readline

        readline.parse_and_bind("tab: complete")
        console.interact()

    @classmethod
    def ipython(cls, local_vars):
        from IPython import start_ipython

        start_ipython(argv=[], user_ns=local_vars)

    @classmethod
    def ptpython(cls, local_vars):
        from ptpython.repl import embed

        embed({}, local_vars)

    @classmethod
    def bpython(cls, local_vars):
        from bpython import embed

        embed(local_vars)

    @classmethod
    def interact(cls, local_vars, preferred_shell=None):
        if preferred_shell:
            shells_to_try = [preferred_shell] + Shell.shells
        else:
            shells_to_try = Shell.shells

        for shell in shells_to_try:
            try:
                return getattr(cls, shell)(local_vars)
            except ImportError:
                pass
            except Exception:
                _logger.error("Could not start '%s' shell.", shell)
                _logger.debug("Shell error:", exc_info=True)

        _logger.error("Could not start any shell.")


def _isatty(stream):
    try:
        return os.isatty(stream.fileno())
    except Exception:
        return False



import sys

env = env  # noqa


op = ""
if len(sys.argv) > 1:
    op = sys.argv[1]

env["ir.config_parameter"].set_param("testparam", "testvalue")
if op == "commit":
    env.cr.commit()
elif op == "rollback":
    env.cr.rollback()
elif op == "raise":
    raise RuntimeError("testerror")
else:
    pass



#!/usr/bin/env python
# Copyright 2018 ACSONE SA/NV (<http://acsone.eu>)
# License LGPL-3.0 or later (http://www.gnu.org/licenses/lgpl.html).

import logging
import runpy
import sys

import click

from . import console
from .env_options import env_options

_logger = logging.getLogger(__name__)


@click.command(
    help="Execute a python script in an initialized Odoo "
    "environment. The script has access to a 'env' global "
    "variable which is an odoo.api.Environment "
    "initialized for the given database. If no script is "
    "provided, the script is read from stdin or an "
    "interactive console is started if stdin appears "
    "to be a terminal."
)
@env_options(database_required=False, with_addons_path=True)
@click.option(
    "--interactive/--no-interactive",
    "-i",
    help="Inspect interactively after running the script.",
)
@click.option(
    "--shell-interface",
    help="Preferred shell interface for interactive mode. Accepted "
    "values are ipython, ptpython, bpython, python. If not "
    "provided they are tried in this order.",
)
@click.argument("script", required=False, type=click.Path(exists=True, dir_okay=False))
@click.argument("script-args", nargs=-1)
def main(env, interactive, shell_interface, script, script_args):
    global_vars = {"env": env}
    if script:
        sys.argv[1:] = script_args
        global_vars = runpy.run_path(
            script, init_globals=global_vars, run_name="__main__"
        )
    if not script or interactive:
        if console._isatty(sys.stdin):
            if not env:
                _logger.info("No environment set, use `-d dbname` to get one.")
            console.Shell.interact(global_vars, shell_interface)
            if env:
                env.cr.rollback()
        else:
            sys.argv[:] = [""]
            global_vars["__name__"] = "__main__"
            exec(sys.stdin.read(), global_vars)


if __name__ == "__main__":
    main()



# Copyright 2018 ACSONE SA/NV (<http://acsone.eu>)
# License LGPL-3.0 or later (http://www.gnu.org/licenses/lgpl.html).

from __future__ import print_function

import os
import subprocess
import sys
import textwrap

import click
import psycopg2
import pytest
from click.testing import CliRunner

import click_odoo
from click_odoo import OdooEnvironment, console, odoo, odoo_bin
from click_odoo.cli import main

here = os.path.abspath(os.path.dirname(__file__))

# This hack is necessary because the way CliRunner patches
# stdout is not compatible with the Odoo logging initialization
# mechanism. Logging is therefore tested with subprocesses.
odoo.netsvc._logger_init = True


def _init_odoo_db(dbname):
    subprocess.check_call(["createdb", dbname])
    subprocess.check_call([odoo_bin, "-d", dbname, "-i", "base", "--stop-after-init"])


def _drop_db(dbname):
    subprocess.check_call(["dropdb", "--if-exists", dbname])


@pytest.fixture(scope="session")
def odoodb():
    if "CLICK_ODOO_TEST_DB" in os.environ:
        yield os.environ["CLICK_ODOO_TEST_DB"]
    else:
        dbname = "click-odoo-test-" + odoo.release.version.replace(".", "-")
        try:
            _init_odoo_db(dbname)
            yield dbname
        finally:
            _drop_db(dbname)


def test_odoo_env(odoodb):
    with OdooEnvironment(database=odoodb) as env:
        admin = env["res.users"].search([("login", "=", "admin")])
        assert len(admin) == 1


def test_click_odoo(odoodb):
    """ Test simple access to env in script """
    script = os.path.join(here, "scripts", "script1.py")
    cmd = ["click-odoo", "-d", odoodb, script]
    result = subprocess.check_output(cmd, universal_newlines=True)
    assert result == "admin\n"


def test_click_odoo_minus_m(odoodb):
    """ Test simple access to env in script """
    script = os.path.join(here, "scripts", "script1.py")
    cmd = [sys.executable, "-m", "click_odoo", "-d", odoodb, script]
    result = subprocess.check_output(cmd, universal_newlines=True)
    assert result == "admin\n"


def test_click_odoo_minus_m_cli(odoodb):
    """ Test simple access to env in script """
    script = os.path.join(here, "scripts", "script1.py")
    cmd = [sys.executable, "-m", "click_odoo.cli", "-d", odoodb, script]
    result = subprocess.check_output(cmd, universal_newlines=True)
    assert result == "admin\n"


def test_cli_runner(odoodb):
    """ Test simple access to env in script (through click CliRunner) """
    script = os.path.join(here, "scripts", "script1.py")
    runner = CliRunner()
    result = runner.invoke(main, ["-d", odoodb, script])
    assert result.exit_code == 0
    assert result.output == "admin\n"


def test_click_odoo_args(odoodb):
    """ Test sys.argv in script """
    script = os.path.join(here, "scripts", "script2.py")
    cmd = ["click-odoo", "-d", odoodb, "--", script, "a", "-b", "-d"]
    result = subprocess.check_output(cmd, universal_newlines=True)
    assert result == textwrap.dedent(
        """\
        sys.argv = {} a -b -d
        __name__ = __main__
    """.format(
            script
        )
    )


def test_click_odoo_shebang(odoodb):
    """ Test simple access to env in script with click-odoo shebang """
    script = os.path.join(here, "scripts", "script1.py")
    cmd = [script, "-d", odoodb]
    result = subprocess.check_output(cmd, universal_newlines=True)
    assert result == "admin\n"


def test_click_odoo_shebang_args(odoodb):
    """ Test script arguments (with click-odoo shebang) """
    script = os.path.join(here, "scripts", "script2.py")
    cmd = [script, "-d", odoodb, "--", "a", "-b", "-d"]
    result = subprocess.check_output(cmd, universal_newlines=True)
    assert result == textwrap.dedent(
        """\
        sys.argv = {} a -b -d
        __name__ = __main__
    """.format(
            script
        )
    )


def test_interactive_no_script(mocker, odoodb):
    mocker.patch.object(console.Shell, "ipython")
    mocker.patch.object(console.Shell, "python")
    mocker.patch.object(console, "_isatty", return_value=True)

    runner = CliRunner()
    result = runner.invoke(main, ["-d", odoodb])
    assert result.exit_code == 0
    assert console.Shell.ipython.call_count == 1
    assert console.Shell.python.call_count == 0


def test_interactive_no_script_preferred_shell(mocker, odoodb):
    mocker.patch.object(console.Shell, "ipython")
    mocker.patch.object(console.Shell, "python")
    mocker.patch.object(console, "_isatty", return_value=True)

    runner = CliRunner()
    result = runner.invoke(main, ["-d", odoodb, "--shell-interface=python"])
    assert result.exit_code == 0
    assert console.Shell.ipython.call_count == 0
    assert console.Shell.python.call_count == 1


def test_logging_stderr(capfd, odoodb):
    script = os.path.join(here, "scripts", "script3.py")
    cmd = ["click-odoo", "-d", odoodb, "--", script]
    subprocess.check_call(cmd)
    out, err = capfd.readouterr()
    assert not out
    assert "Modules loaded" in err
    assert "hello from script3" in err


def test_logging_logfile(tmpdir, capfd, odoodb):
    script = os.path.join(here, "scripts", "script3.py")
    logfile = tmpdir.join("mylogfile")
    cmd = ["click-odoo", "-d", odoodb, "--logfile", str(logfile), "--", script]
    subprocess.check_call(cmd)
    out, err = capfd.readouterr()
    assert not out
    logcontent = logfile.read()
    assert "Modules loaded" in logcontent
    assert "hello from script3" in logcontent


def test_env_options_withdb(odoodb, tmpdir):
    @click.command()
    @click_odoo.env_options()
    def testcmd(env):
        login = env["res.users"].search([("login", "=", "admin")]).login
        click.echo("login={}".format(login))

    # database from command line
    runner = CliRunner()
    result = runner.invoke(testcmd, ["-d", odoodb])
    assert result.exit_code == 0
    assert "login=admin\n" in result.output
    # database in config
    odoocfg1 = tmpdir / "odoo1.cfg"
    odoocfg1.write(
        textwrap.dedent(
            """\
        [options]
        db_name={}
    """.format(
                odoodb
            )
        )
    )
    result = runner.invoke(testcmd, ["-c", str(odoocfg1)])
    assert result.exit_code == 0
    assert "login=admin\n" in result.output
    # database -d has priority over db_name in config
    odoocfg2 = tmpdir / "odoo2.cfg"
    odoocfg2.write(
        textwrap.dedent(
            """\
        [options]
        db_name=notadb
    """
        )
    )
    result = runner.invoke(testcmd, ["-c", str(odoocfg2), "-d", odoodb])
    assert result.exit_code == 0
    assert "login=admin\n" in result.output
    # test multiple databases in config file
    odoocfg3 = tmpdir / "odoo3.cfg"
    odoocfg3.write(
        textwrap.dedent(
            """\
        [options]
        db_name={},notadb
    """.format(
                odoodb
            )
        )
    )
    result = runner.invoke(testcmd, ["-c", str(odoocfg3)])
    assert result.exit_code != 0
    assert "No database provided" in result.output
    result = runner.invoke(testcmd, ["-c", str(odoocfg3), "-d", odoodb])
    assert result.exit_code == 0
    assert "login=admin\n" in result.output
    # no -d, error
    result = runner.invoke(testcmd, [])
    assert result.exit_code != 0
    assert "No database provided" in result.output


def test_env_options_nodb(odoodb, tmpdir):
    @click.command()
    @click_odoo.env_options(with_database=False)
    def testcmd(env):
        assert not env

    # no database
    runner = CliRunner()
    result = runner.invoke(testcmd, [])
    assert result.exit_code == 0
    # -d not allowed
    result = runner.invoke(testcmd, ["-d", odoodb])
    assert result.exit_code != 0
    assert "no such option: -d" in result.output
    # db_name in config ignored
    odoocfg1 = tmpdir / "odoo1.cfg"
    odoocfg1.write(
        textwrap.dedent(
            """\
        [options]
        db_name={}
    """.format(
                odoodb
            )
        )
    )
    result = runner.invoke(testcmd, ["-c", str(odoocfg1)])
    assert result.exit_code == 0


def test_env_options_optionaldb(odoodb, tmpdir):
    @click.command()
    @click_odoo.env_options(database_required=False)
    def testcmd(env):
        if env:
            print("with env")
        else:
            print("without env")

    # no database
    runner = CliRunner()
    result = runner.invoke(testcmd, [])
    assert result.exit_code == 0
    assert "without env" in result.output
    # with database
    runner = CliRunner()
    result = runner.invoke(testcmd, ["-d", odoodb])
    assert result.exit_code == 0
    assert "with env" in result.output
    # database in config
    odoocfg1 = tmpdir / "odoo1.cfg"
    odoocfg1.write(
        textwrap.dedent(
            """\
        [options]
        db_name={}
    """.format(
                odoodb
            )
        )
    )
    result = runner.invoke(testcmd, ["-c", str(odoocfg1)])
    assert result.exit_code == 0
    assert "with env" in result.output


def test_env_options_database_must_exist(odoodb):
    @click.command()
    @click_odoo.env_options(database_must_exist=False)
    def testcmd(env):
        if env:
            print("with env")
        else:
            print("without env")

    # no database, must not exist, no env
    runner = CliRunner()
    result = runner.invoke(testcmd, ["-d", "dbthatdoesnotexist"])
    assert result.exit_code == 0
    assert "without env" in result.output

    # database exists, must not exist, env ok
    runner = CliRunner()
    result = runner.invoke(testcmd, ["-d", odoodb])
    assert result.exit_code == 0
    assert "with env" in result.output


def _cleanup_testparam(dbname):
    with psycopg2.connect(dbname=dbname) as conn:
        with conn.cursor() as cr:
            cr.execute("DELETE FROM ir_config_parameter " "WHERE key='testparam'")
            conn.commit()
    conn.close()


def _assert_testparam_present(dbname, expected):
    with psycopg2.connect(dbname=dbname) as conn:
        with conn.cursor() as cr:
            cr.execute("SELECT value FROM ir_config_parameter " "WHERE key='testparam'")
            r = cr.fetchall()
            assert len(r) == 1
            assert r[0][0] == expected
    conn.close()


def _assert_testparam_absent(dbname):
    with psycopg2.connect(dbname=dbname) as conn:
        with conn.cursor() as cr:
            cr.execute("SELECT value FROM ir_config_parameter " "WHERE key='testparam'")
            r = cr.fetchall()
            assert len(r) == 0
    conn.close()


def test_write_commit(odoodb):
    """ test commit in script """
    _cleanup_testparam(odoodb)
    script = os.path.join(here, "scripts", "script4.py")
    cmd = ["click-odoo", "-d", odoodb, "--", script, "commit"]
    subprocess.check_call(cmd)
    _assert_testparam_present(odoodb, "testvalue")


def test_write_rollback(odoodb):
    """ test rollback in script """
    _cleanup_testparam(odoodb)
    script = os.path.join(here, "scripts", "script4.py")
    cmd = ["click-odoo", "-d", odoodb, "--", script, "rollback"]
    subprocess.check_call(cmd)
    _assert_testparam_absent(odoodb)


def test_write_defaulttx(odoodb):
    """ test click-odoo commits itself """
    _cleanup_testparam(odoodb)
    script = os.path.join(here, "scripts", "script4.py")
    cmd = ["click-odoo", "-d", odoodb, "--", script]
    subprocess.check_call(cmd)
    _assert_testparam_present(odoodb, "testvalue")


def test_write_interactive_defaulttx(mocker, odoodb):
    """ test click-odoo rollbacks in interactive mode """
    mocker.patch.object(console.Shell, "python")
    mocker.patch.object(console, "_isatty", return_value=True)

    _cleanup_testparam(odoodb)
    runner = CliRunner()
    script = os.path.join(here, "scripts", "script4.py")
    cmd = ["-d", odoodb, "--interactive", "--", script]
    result = runner.invoke(main, cmd)
    assert result.exit_code == 0
    _assert_testparam_absent(odoodb)


def test_write_stdin_defaulttx(odoodb):
    _cleanup_testparam(odoodb)
    script = os.path.join(here, "scripts", "script4.py")
    cmd = ["click-odoo", "-d", odoodb, "<", script]
    subprocess.check_call(" ".join(cmd), shell=True)
    _assert_testparam_present(odoodb, "testvalue")


def test_write_raise(tmpdir, capfd, odoodb):
    """ test nothing is committed if the script raises """
    _cleanup_testparam(odoodb)
    script = os.path.join(here, "scripts", "script4.py")
    logfile = tmpdir.join("mylogfile")
    cmd = ["click-odoo", "-d", odoodb, "--logfile", str(logfile), "--", script, "raise"]
    r = subprocess.call(cmd)
    assert r != 0
    logcontent = logfile.read()
    assert "testerror" in logcontent
    out, err = capfd.readouterr()
    assert "testerror" in err
    _assert_testparam_absent(odoodb)


def test_env_cache(odoodb):
    """ test a new environment does not reuse cache """
    _cleanup_testparam(odoodb)
    with OdooEnvironment(database=odoodb) as env:
        env["ir.config_parameter"].set_param("testparam", "testvalue")
        value = env["ir.config_parameter"].get_param("testparam")
        assert value == "testvalue"
        env.cr.commit()
    _assert_testparam_present(odoodb, "testvalue")
    _cleanup_testparam(odoodb)
    _assert_testparam_absent(odoodb)
    with OdooEnvironment(database=odoodb) as env:
        value = env["ir.config_parameter"].get_param("testparam")
        assert not value


def test_env_options_addons_path():
    script = os.path.join(here, "scripts", "script5.py")

    cmd = ["click-odoo", "--", script]
    r = subprocess.call(cmd)
    assert r != 0  # addon1 not found in addons path

    addons_path = ",".join(
        [
            os.path.join(odoo.__path__[0], "addons"),
            os.path.join(os.path.dirname(__file__), "data", "addons"),
        ]
    )

    cmd = ["click-odoo", "--addons-path", addons_path, "--", script]
    r = subprocess.call(cmd)
    assert r == 0


def test_parse_version():
    versions = (
        (
            "4.2",
            "4.2.3.4",
            "5.0.0-alpha",
            "5.0.0-rc1",
            "5.0.0-rc1.1",
            "5.0.0_rc2",
            "5.0.0_rc3",
            "5.0.0",
            "12.0",
            "saas~12.5",
            "13.0",
        ),
        ("5.0.0-0_rc3", "5.0.0-1dev", "5.0.0-1"),
    )
    for sequence in versions:
        smaller = "0"
        for bigger in sequence:
            assert odoo.tools.parse_version(smaller) < odoo.tools.parse_version(bigger)
            smaller = bigger



#!/usr/bin/env python
import os
import subprocess
import sys

odoo_branch = sys.argv[1]
odoo_dir = sys.argv[2]


def odoo_installed():
    try:
        import odoo  # noqa

        return True
    except ImportError:
        # odoo < 10
        try:
            import openerp  # noqa

            return True
        except ImportError:
            # odoo not installed
            return False


def odoo_cloned():
    return os.path.isdir(os.path.join(odoo_dir, ".git"))


def clone_odoo():
    subprocess.check_call(
        [
            "git",
            "clone",
            "--depth=1",
            "-b",
            odoo_branch,
            "https://github.com/odoo/odoo",
            odoo_dir,
        ]
    )


def install_odoo():
    # Odoo not compatible with recent pyyaml due to load() now being safe by default
    # XXX an alternative is to pip install -r requirements.txt with
    # XXX the correct one for each Odoo version but this is slower as
    # XXX many wheels more need to be built because Odoo's requirements.txt
    # XXX pin old library versions.
    subprocess.check_call(["pip", "install", "pyyaml<4"])
    subprocess.check_call(["pip", "install", "-e", odoo_dir])


def main():
    if not odoo_installed():
        if not odoo_cloned():
            clone_odoo()
        install_odoo()


if __name__ == "__main__":
    main()



# noqa
try:
    import odoo.release as release
    import odoo.tools.parse_version as parse_version
except (AttributeError, ImportError):
    import openerp.release as release
    import openerp.tools.parse_version as parse_version

if parse_version(release.version) > parse_version("9.0c"):
    import odoo.addons.addon1  # noqa
else:
    import openerp.addons.addon1  # noqa



# Copyright 2018 ACSONE SA/NV (<http://acsone.eu>)
# License LGPL-3.0 or later (http://www.gnu.org/licenses/lgpl.html).

import os

from setuptools import setup

here = os.path.abspath(os.path.dirname(__file__))

long_description = []
with open(os.path.join("README.rst")) as f:
    long_description.append(f.read())
with open(os.path.join("CHANGES.rst")) as f:
    long_description.append(f.read())


setup(
    name="click-odoo",
    description="Beautiful, robust CLI for Odoo",
    long_description="\n".join(long_description),
    use_scm_version=True,
    packages=["click_odoo"],
    include_package_data=True,
    setup_requires=["setuptools_scm"],
    install_requires=["click>=7"],
    license="LGPLv3+",
    author="ACSONE SA/NV",
    author_email="info@acsone.eu",
    url="http://github.com/acsone/click-odoo",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: "
        "GNU Lesser General Public License v3 or later (LGPLv3+)",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Framework :: Odoo",
    ],
    entry_points="""
        [console_scripts]
        click-odoo=click_odoo.cli:main
    """,
)



# Copyright 2018 ACSONE SA/NV (<http://acsone.eu>)
# License LGPL-3.0 or later (http://www.gnu.org/licenses/lgpl.html).

import logging
import sys
from contextlib import closing

import click
from click.decorators import _param_memo  # XXX undocumented click internal

from .env import OdooEnvironment, odoo

_logger = logging.getLogger(__name__)


class env_options(object):
    def __init__(
        self,
        default_log_level="info",
        with_rollback=True,
        with_database=True,
        database_required=True,
        database_must_exist=True,
        with_addons_path=False,
        environment_manager=OdooEnvironment,
    ):
        self.default_log_level = default_log_level
        self.with_rollback = with_rollback
        self.with_database = with_database
        self.database_required = database_required
        self.database_must_exist = database_must_exist
        self.with_addons_path = with_addons_path
        self.environment_manager = environment_manager

    def __call__(self, f):
        # this is the decorator call which registers options in reverse order
        if self.with_rollback:
            _param_memo(
                f,
                click.Option(
                    ("--rollback",),
                    is_flag=True,
                    help="Rollback the transaction even if the script "
                    "does not raise an exception. Note that if the "
                    "script itself commits, this option has no effect. "
                    "This is why it is not named dry run. This option "
                    "is implied when an interactive console is "
                    "started.",
                ),
            )
        _param_memo(
            f,
            click.Option(
                ("--logfile",),
                type=click.Path(dir_okay=False),
                help="Specify the log file.",
            ),
        )
        _param_memo(
            f,
            click.Option(
                ("--log-level",),
                default=self.default_log_level,
                show_default=True,
                help="Specify the logging level. Accepted values depend "
                "on the Odoo version, and include debug, info, "
                "warn, error.",
            ),
        )
        if self.with_database:
            _param_memo(
                f,
                click.Option(
                    ("--database", "-d"),
                    envvar=["PGDATABASE"],
                    help="Specify the database name. If present, this "
                    "parameter takes precedence over the database "
                    "provided in the Odoo configuration file.",
                ),
            )
        if self.with_addons_path:
            _param_memo(
                f,
                click.Option(
                    ("--addons-path",),
                    envvar=["ODOO_ADDONS_PATH"],
                    help="Specify the addons path. If present, this "
                    "parameter takes precedence over the addons path "
                    "provided in the Odoo configuration file.",
                ),
            )
        _param_memo(
            f,
            click.Option(
                ("--config", "-c"),
                envvar=["ODOO_RC", "OPENERP_SERVER"],
                type=click.Path(exists=True, dir_okay=False),
                callback=self._register,
                help="Specify the Odoo configuration file. Other "
                "ways to provide it are with the ODOO_RC or "
                "OPENERP_SERVER environment variables, "
                "or ~/.odoorc (Odoo >= 10) "
                "or ~/.openerp_serverrc.",
            ),
        )
        return f

    def _register(self, ctx, param, value):
        if ctx.command.invoke != self._invoke:
            self.org_invoke = ctx.command.invoke
            ctx.command.invoke = self._invoke
        return value

    def _fix_odoo_logging(self):
        if odoo.tools.parse_version(odoo.release.version) < odoo.tools.parse_version(
            "9.0c"
        ):
            handlers = logging.getLogger().handlers
            if handlers and len(handlers) == 1:
                handler = handlers[0]
                if isinstance(handler, logging.StreamHandler):
                    if handler.stream is sys.stdout:
                        handler.stream = sys.stderr

    def _fix_disable_wsgi_module_handlers(self):
        if odoo.tools.parse_version(odoo.release.version) < odoo.tools.parse_version(
            "9.0c"
        ):
            odoo.service.wsgi_server.module_handlers[:] = []

    def get_odoo_args(self, ctx):
        """Return a list of Odoo command line arguments from the Click context."""
        config = ctx.params.get("config")
        addons_path = ctx.params.get("addons_path")
        database = ctx.params.get("database")
        log_level = ctx.params.get("log_level")
        logfile = ctx.params.get("logfile")

        odoo_args = []

        if config:
            odoo_args.extend(["--config", config])
        if addons_path:
            odoo_args.extend(["--addons-path", addons_path])
        if database:
            odoo_args.extend(["--database", database])
        if log_level:
            odoo_args.extend(["--log-level", log_level])
        if logfile:
            odoo_args.extend(["--logfile", logfile])

        return odoo_args

    def _configure_odoo(self, ctx):
        odoo_args = self.get_odoo_args(ctx)
        # reset db_name in case we come from a previous run
        # where database has been set, in the second run there is no database
        # (mostly for tests)
        odoo.tools.config["db_name"] = None
        # see https://github.com/odoo/odoo/commit/b122217f74
        odoo.tools.config["load_language"] = None
        odoo.tools.config.parse_config(odoo_args)
        self._fix_odoo_logging()
        self._fix_disable_wsgi_module_handlers()
        odoo.cli.server.report_configuration()

    def _db_exists(self, dbname):
        conn = odoo.sql_db.db_connect("postgres")
        with closing(conn.cursor()) as cr:
            cr._obj.execute(
                "SELECT datname FROM pg_catalog.pg_database "
                "WHERE lower(datname) = lower(%s)",
                (dbname,),
            )
            return bool(cr.fetchone())

    def _pop_params(self, ctx):
        ctx.params.pop("config", None)
        ctx.params.pop("addons_path", None)
        ctx.params.pop("database", None)
        ctx.params.pop("log_level", None)
        ctx.params.pop("logfile", None)
        ctx.params.pop("rollback", None)

    def _invoke(self, ctx):
        try:
            self._configure_odoo(ctx)
            database = ctx.params.get("database")
            if (
                not database
                and odoo.tools.config["db_name"]
                and "," not in odoo.tools.config["db_name"]
            ):
                # if database not provided as parameter and a single
                # database is provided in the odoo config file, use it
                database = odoo.tools.config["db_name"]
            rollback = ctx.params.get("rollback")
            # pop env_options params so they are not passed to the command
            self._pop_params(ctx)
            if self.with_database and self.database_required and not database:
                raise click.UsageError(
                    "No database provided, please provide one with the -d "
                    "option or the Odoo configuration file."
                )
            if (
                self.with_database
                and database
                and (self.database_must_exist or self._db_exists(database))
            ):
                with self.environment_manager(
                    database=database, rollback=rollback, ctx=ctx
                ) as env:
                    ctx.params["env"] = env
                    return self.org_invoke(ctx)
            else:
                with odoo.api.Environment.manage():
                    ctx.params["env"] = None
                    return self.org_invoke(ctx)
        except click.exceptions.Exit:
            raise
        except Exception as e:
            _logger.error("exception", exc_info=True)
            raise click.ClickException(str(e))
